
#ifndef DISSECTOR_API_H
#define DISSECTOR_API_H
#include <stdint.h>
#include <stdbool.h>

//suricata autoconf complains about these types when pcap.h is included. it should be in sys/types.h
#ifndef __u_char_defined
typedef __u_char u_char;
typedef __u_short u_short;
typedef __u_int u_int;
typedef __u_long u_long;
#define __u_char_defined
#endif// __u_char_defined

#include <pcap.h>
#ifdef __cplusplus
extern "C" {
#endif

enum {
    L2_FLAG_NULL  = '\0',
    L2_FLAG_VNTAG = 'T',
    L2_FLAG_VLAN  ='V',
    L2_FLAG_MPLS  = 'M'
};

#define L2_FLAGS 16
typedef struct l2_flags_s 
{
    uint8_t flags[L2_FLAGS];
    uint8_t num_flags;//put to end so that flags are aligned.
} l2_flags_t;

#define L2_FLAGS_IS_EMPTY(f) ((f).num_flags == 0)
#define L2_FLAGS_STARTS_WITH(l2_flags, flag) ((l2_flags).num_flags > 0 && (l2_flags).flags[0] == flag)
#define L2_FLAGS_IS_EQUAL(l,r) ((l).num_flags == (r).num_flags && 0 == memcmp((l).flags, (r).flags, (l).num_flags))
 
typedef struct extended_l2_s {
    l2_flags_t flags;
    struct bpf_program bpf_prog;
} extended_l2_t;

enum {
    MULTI_BPF_PROG_INNERMOST = 1,
};

typedef struct multi_bpf_prog_s {
    int32_t flags;
    bool matches_anything;
    char *bpf_filter;
    struct bpf_program bpf_prog_l2;
    struct bpf_program bpf_prog_l3;
    extended_l2_t *xl2s;
    int xl2s_size;
} multi_bpf_prog_t;

/*
 * \brief Get the inner frame information
 *
 * The function assumes the incoming packet is an ethernet frame.
 * 
 * \param pkt [input] Pointer to the packet 
 * \param len [input] Length of the packet
 * \param l2_offset [output] the layer 2 offset of the innermost ethernet frame.
 * \param l2_flags [output] indicates l2 encapsulation layers of the innermost ethernet: vntag, vlan or mpls. This is used to form special bpfs.
 * \param l3_offset [output] indicates the l3 layer of the innermost ethernet frame which should point to IPv4 or IPv6 header. it's the same as l2_offset when no such frame exists.
 *
 * \return false if the packet is detected incomplete for parsing. true if parsed successfully
 * 
 */

bool parse_inner_frame_info(uint8_t *pkt, uint32_t len, uint32_t *l2_offset, l2_flags_t *l2_flags, uint32_t *l3_offset);

/*
 * \brief Given the base_filter, generatates extended filter by adding vlan and / or mpls before ether 
 * when vlan or/and mpls are detected in the ether frame. This way the base filter doesn't have to care about mpls or vlan encapsulations right after ether.
 * 
*/

bool get_extended_filter(l2_flags_t l2_flags, const char* base_filter, char* extended_filter, uint32_t extended_filter_size);

/*
 * in case get_extended_filter returns false, that means extended_filter_size is not big enough. call this function to get the required size.
*/ 
uint32_t get_extended_filter_size(l2_flags_t l2_flags, const char* base_filter);

/*
 * \brief initialize multi_bpf_prog
 * \param flags indicate wether to apply it on the whole packet or the innermost frame.
*/ 
void multi_bpf_prog_init(multi_bpf_prog_t* multi_bpf_prog, uint32_t flags);

/*
 * \brief compile a multi_bpf_prog from basic bpf_filter 
 * multi_bpf_prog is a set of bpf programs to handle packets with encapsulation layers effective.
 * in this scheme, basic_bpf_filter is the basic bpf filter not containing any encapsulation filters such as vlan or mpls.
 * based on the basic_bpf_filter and the packet, it applies the right bpf program to the right frame of the packet.
 * 
 * \param multi_bpf_prog contains a set of bpf programs.
 * \param snaplen_arg 
 * \param basic_bpf_filter the base filter to be applied to the packet
 * \param errbuf to store error in case of failure.
 * \param errbuf_len length of the errbuf.
 *
 * \retval 0 success, -1 failure.
 */

bool multi_bpf_prog_compile(multi_bpf_prog_t* multi_bpf_prog, int snaplen_arg, const char *base_bpf_filter, char *errbuf, size_t errbuf_len);

/*
 * \brief use multi_bpf_prog to filter packet
 */

bool multi_bpf_prog_filter(multi_bpf_prog_t* multi_bpf_prog, uint8_t *packet, uint32_t packet_len, uint32_t caplen);

/*
 * \brief free multi_bpf_prog
 */

void multi_bpf_prog_free(multi_bpf_prog_t* multi_bpf_prog);

/*
 * \brief compile a single bpf program
 */

bool single_bpf_prog_compile(struct bpf_program *program, int snaplen_arg, int linktype_arg, const char *buf, int optimize, uint32_t mask, char *errbuf, size_t errbuf_len);

/*
 * \brief filter packet with single bpf program.
 */

bool single_bpf_prog_filter(struct bpf_program* bpf_prog, uint8_t *packet, uint32_t offset, uint32_t packet_len, uint32_t caplen);

/*
 * \brief get l2_flags type from an array of flags with NULL at the end.
 */

l2_flags_t get_l2_flags(uint8_t* flags);

#ifdef __cplusplus
}
#endif

#endif //DISSECTOR_API_H